from __future__ import annotations

import asyncio
import csv
import gzip
import io
import sys
from typing import AsyncIterator, Iterable, Sequence, TypedDict

import psycopg
from psycopg import sql as psql
from psycopg.rows import dict_row
from psycopg_pool import AsyncConnectionPool

from .sql import (
    TABLE_PRESETS,
)

_CHUNK = 1024 * 1024  # 1MB chunks for streaming


def _open_maybe_gz_write(path: str):
    """Return a binary file-like for writing (stdout if '-')."""
    if path == "-":
        # write to stdout (binary)
        return sys.stdout.buffer, False  # (fh, close_when_done)
    if path.endswith(".gz"):
        return gzip.open(path, "wb"), True
    return open(path, "wb"), True


def _open_maybe_gz_read_text(path: str):
    """Return a text file-like for reading CSV/NDJSON (stdin if '-')."""
    if path == "-":
        # don't close user's stdin
        return io.TextIOWrapper(sys.stdin.buffer, encoding="utf-8"), False
    if path.endswith(".gz"):
        return gzip.open(path, "rt", encoding="utf-8"), True
    return open(path, "rt", encoding="utf-8"), True


async def _aiter_text_chunks(fh: io.TextIOBase, size: int = _CHUNK) -> AsyncIterator[str]:
    """Async generator yielding text chunks from a file-like using a thread offload."""
    while True:
        chunk = await asyncio.to_thread(fh.read, size)
        if not chunk:
            break
        yield chunk


def upsert_statement(
    table: str,
    cols: Sequence[str],
    conflict_cols: Sequence[str],
    update_cols: Sequence[str],
) -> psql.Composed:
    """INSERT ... ON CONFLICT ... DO UPDATE with named parameters (%(name)s)."""
    ins_cols = psql.SQL(", ").join(psql.Identifier(c) for c in cols)
    ins_vals = psql.SQL(", ").join(psql.Placeholder(c) for c in cols)
    conflict = psql.SQL(", ").join(psql.Identifier(c) for c in conflict_cols)
    setlist = psql.SQL(", ").join(
        psql.SQL("{} = EXCLUDED.{}").format(psql.Identifier(c), psql.Identifier(c))
        for c in update_cols
    )
    return psql.SQL("INSERT INTO {} ({}) VALUES ({}) ON CONFLICT ({}) DO UPDATE SET {}").format(
        psql.Identifier(table), ins_cols, ins_vals, conflict, setlist
    )


def latest_prices_select(symbols: Iterable[str], vendor: str, tenant_id: str) -> psql.Composed:
    # Uses your view latest_prices(tenant_id,vendor,symbol,price,price_timestamp)
    return psql.SQL(
        "SELECT vendor, symbol, price, price_timestamp "
        "FROM latest_prices WHERE tenant_id = {tid} AND vendor = {v} AND symbol = ANY({syms})"
    ).format(
        tid=psql.Literal(tenant_id),
        v=psql.Literal(vendor),
        syms=psql.Literal(list({s.upper() for s in symbols})),
    )


def bars_window_select(
    *, symbol: str, timeframe: str, start: str, end: str, vendor: str
) -> psql.Composed:
    return psql.SQL(
        "SELECT ts, tenant_id, vendor, symbol, timeframe, open_price, high_price, "
        "low_price, close_price, volume "
        "FROM bars "
        "WHERE vendor = {v} AND symbol = {s} AND timeframe = {tf} "
        "AND ts >= {start} AND ts < {end} "
        "ORDER BY ts"
    ).format(
        v=psql.Literal(vendor),
        s=psql.Literal(symbol.upper()),
        tf=psql.Literal(timeframe),
        start=psql.Literal(start),
        end=psql.Literal(end),
    )


def copy_to_stdout_ndjson(select_json_sql: psql.Composed) -> psql.Composed:
    # Expect a SELECT producing a single json/jsonb column per row.
    return psql.SQL("COPY ({}) TO STDOUT").format(select_json_sql)


def copy_to_stdout_csv(select_sql: psql.Composed) -> psql.Composed:
    return psql.SQL("COPY ({}) TO STDOUT WITH CSV HEADER").format(select_sql)


class AMDSConfig(TypedDict, total=False):
    dsn: str
    tenant_id: str
    app_name: str
    statement_timeout_ms: int
    pool_max: int
    write_mode: str  # "auto" | "executemany" | "copy"   (async: no execute_values)
    copy_min_rows: int


DEFAULTS: AMDSConfig = {
    "pool_max": 10,
    "write_mode": "auto",
    "copy_min_rows": 5000,
}


class AMDS:
    def __init__(self, cfg: AMDSConfig):
        self.cfg: AMDSConfig = {**DEFAULTS, **(cfg or {})}
        if "dsn" not in self.cfg:
            raise ValueError("dsn required")
        # Create pool without auto-opening (fixes deprecation warning)
        self.pool = AsyncConnectionPool(
            conninfo=self.cfg["dsn"],
            max_size=self.cfg["pool_max"],
            kwargs={"autocommit": False},
            open=False,  # Never auto-open in constructor
        )
        self._connection_preparator = self._prepare_async_conn
        self.tenant_id = self.cfg.get("tenant_id")
        self.statement_timeout_ms = self.cfg.get("statement_timeout_ms")
        self.app_name = self.cfg.get("app_name")
        self._pool_opened = False

    async def __aenter__(self):
        await self.aopen()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.aclose()

    async def aopen(self) -> None:
        """Explicitly open the connection pool."""
        if not self._pool_opened:
            await self.pool.open(wait=True, timeout=5.0)
            self._pool_opened = True

    async def aclose(self) -> None:
        """Close the connection pool with proper cleanup."""
        if self._pool_opened:
            from .runtime import shutdown_with_timeout

            await shutdown_with_timeout(self.pool, timeout=1.0)
            self._pool_opened = False

    async def _prepare_async_conn(self, conn):
        """Prepare connection with app name and timeouts."""
        # Note: app.tenant_id parameter not supported in this database
        # Tenant isolation is handled via RLS policies instead
        if self.app_name:
            await conn.execute("SET application_name = %s", (self.app_name,))
        if self.statement_timeout_ms:
            await conn.execute("SET statement_timeout = %s", (self.statement_timeout_ms,))

    async def _conn(self):
        """Get connection with pre-configured tenant, app name, and timeouts."""
        # Ensure pool is opened
        await self.aopen()

        async with self.pool.connection() as conn:
            # Apply connection preparation
            await self._prepare_async_conn(conn)
            yield conn

    # ---------- health / meta ----------

    async def health(self) -> bool:
        async for conn in self._conn():
            await conn.execute("SELECT 1")
            return True

    async def schema_version(self) -> str | None:
        async for conn in self._conn():
            try:
                cur = await conn.execute("SELECT version_num FROM alembic_version LIMIT 1")
                row = await cur.fetchone()
                return row[0] if row else None
            except psycopg.errors.UndefinedTable:
                return None

    # ---------- generic upsert ----------

    def _coerce_rows(self, rows: Iterable[object]) -> list[dict]:
        out: list[dict] = []
        for r in rows:
            if r is None:
                continue
            if hasattr(r, "model_dump"):
                out.append(r.model_dump(exclude_none=True))
            elif isinstance(r, dict):
                out.append({k: v for k, v in r.items() if v is not None})
            else:
                out.append({k: v for k, v in vars(r).items() if v is not None})
        return out

    def _write_mode(self, nrows: int) -> str:
        mode = (self.cfg.get("write_mode") or "auto").lower()
        if mode != "auto":
            return mode
        if nrows >= int(self.cfg["copy_min_rows"]):
            return "copy"
        return "executemany"

    async def _copy_from_memory_csv(
        self, conn: psycopg.AsyncConnection, table: str, cols: Sequence[str], rows: Sequence[dict]
    ):
        sio = io.StringIO()
        writer = csv.DictWriter(sio, fieldnames=list(cols))
        writer.writeheader()
        for r in rows:
            writer.writerow({c: r.get(c) for c in cols})
        sio.seek(0)
        async with (
            conn.cursor() as cur,
            cur.copy(
                psql.SQL("COPY {} ({}) FROM STDIN WITH CSV HEADER").format(
                    psql.Identifier(table),
                    psql.SQL(", ").join(psql.Identifier(c) for c in cols),
                )
            ) as cp,
        ):
            await cp.write(sio.read())

    async def _upsert(self, table: str, rows: Iterable[object]) -> int:
        preset = TABLE_PRESETS[table]
        cols, conflict, update = preset.cols, preset.conflict, preset.update
        sql_stmt = upsert_statement(table, cols, conflict, update)
        data = self._coerce_rows(rows)
        if not data:
            return 0

        async for conn in self._conn():
            async with conn.cursor(row_factory=dict_row) as cur:
                mode = self._write_mode(len(data))
                if mode == "executemany":
                    await cur.executemany(sql_stmt, data)
                elif mode == "copy":
                    temp = psql.Identifier(f"tmp_{table}_copy")
                    await cur.execute(
                        psql.SQL(
                            "CREATE TEMP TABLE {} (LIKE {} INCLUDING DEFAULTS) ON COMMIT DROP"
                        ).format(temp, psql.Identifier(table))
                    )
                    await self._copy_from_memory_csv(conn, temp.string, cols, data)
                    ins = psql.SQL(
                        "INSERT INTO {} ({cols}) SELECT {cols} FROM {} "
                        "ON CONFLICT ({conf}) DO UPDATE SET {upd}"
                    ).format(
                        psql.Identifier(table),
                        temp,
                        cols=psql.SQL(", ").join(psql.Identifier(c) for c in cols),
                        conf=psql.SQL(", ").join(psql.Identifier(c) for c in conflict),
                        upd=psql.SQL(", ").join(
                            psql.SQL("{} = EXCLUDED.{}").format(
                                psql.Identifier(c), psql.Identifier(c)
                            )
                            for c in update
                        ),
                    )
                    await cur.execute(ins)
                else:
                    raise ValueError(f"unknown write_mode {mode}")
            await conn.commit()
        return len(data)

    # ---------- typed upserts ----------

    async def upsert_bars(self, rows: Sequence[object]) -> int:
        return await self._upsert("bars", rows)

    async def upsert_fundamentals(self, rows: Sequence[object]) -> int:
        return await self._upsert("fundamentals", rows)

    async def upsert_news(self, rows: Sequence[object]) -> int:
        return await self._upsert("news", rows)

    async def upsert_options(self, rows: Sequence[object]) -> int:
        return await self._upsert("options_snap", rows)

    # ---------- reads ----------

    async def latest_prices(self, symbols: Iterable[str], vendor: str) -> list[dict]:
        if not self.tenant_id:
            raise ValueError("tenant_id required for latest_prices()")
        q = latest_prices_select(symbols, vendor, self.tenant_id)
        async for conn in self._conn():
            async with conn.cursor(row_factory=dict_row) as cur:
                await cur.execute(q)
                return list(await cur.fetchall())

    async def bars_window(
        self, *, symbol: str, timeframe: str, start: str, end: str, vendor: str
    ) -> list[dict]:
        q = bars_window_select(
            symbol=symbol, timeframe=timeframe, start=start, end=end, vendor=vendor
        )
        async for conn in self._conn():
            async with conn.cursor(row_factory=dict_row) as cur:
                await cur.execute(q)
                return list(await cur.fetchall())

    # ---------- COPY export (CSV / NDJSON) ----------

    async def copy_out_csv(self, *, select_sql: psql.Composed, out_path: str) -> int:
        copy_sql = copy_to_stdout_csv(select_sql)
        writer = gzip.open(out_path, "wb") if out_path.endswith(".gz") else open(out_path, "wb")
        try:
            async for conn in self._conn():
                async with conn.cursor() as cur, cur.copy(copy_sql) as cp:
                    n = 0
                    while True:
                        chunk = await cp.read()
                        if not chunk:
                            break
                        writer.write(chunk)
                        n += len(chunk)
                    return n
        finally:
            writer.close()

    async def copy_out_ndjson_async(self, *, select_sql: psql.SQL, out_path: str) -> int:
        """
        COPY (SELECT to_jsonb(...)) TO STDOUT into NDJSON file (or stdout if '-').
        Returns the total bytes written. Gzip supported via *.gz.
        """
        total = 0
        fh, should_close = _open_maybe_gz_write(out_path)
        try:
            async with self.pool.connection() as conn:
                async with conn.cursor() as cur:
                    copy_sql = psql.SQL("COPY ({sel}) TO STDOUT").format(sel=select_sql)
                    async with cur.copy(copy_sql) as cp:
                        # cp.read() yields bytes (server text stream)
                        while True:
                            data = await cp.read()
                            if not data:
                                break
                            # write bytes without decoding
                            await asyncio.to_thread(fh.write, data)
                            total += len(data)
            # flush to disk if needed
            await asyncio.to_thread(fh.flush)
        finally:
            if should_close:
                try:
                    fh.close()
                except Exception:
                    pass
        return total

    async def copy_restore_csv_async(
        self,
        *,
        target: str,
        cols: Sequence[str],
        conflict_cols: Sequence[str],
        update_cols: Sequence[str],
        src_path: str,
        csv_has_header: bool = True,
        csv_delimiter: str = ",",
    ) -> int:
        """
        Restore CSV (optionally .gz or stdin '-') using a TEMP staging table,
        then INSERT ... ON CONFLICT DO UPDATE into {target}.
        Returns affected row count.
        """
        col_idents = psql.SQL(", ").join(psql.Identifier(c) for c in cols)
        conflict_idents = psql.SQL(", ").join(psql.Identifier(c) for c in conflict_cols)
        set_list = psql.SQL(", ").join(
            psql.SQL("{c}=EXCLUDED.{c}").format(c=psql.Identifier(c)) for c in update_cols
        )
        tmp = psql.Identifier(f"_staging_{target}")

        async with self.pool.connection() as conn:
            async with conn.cursor() as cur:
                # Create staging table (inherits defaults for consistent types)
                await cur.execute(
                    psql.SQL("CREATE TEMP TABLE {tmp} (LIKE {t} INCLUDING DEFAULTS)").format(
                        tmp=tmp, t=psql.Identifier(target)
                    )
                )

                copy_sql = psql.SQL(
                    "COPY {tmp} ({cols}) FROM STDIN WITH (FORMAT csv, HEADER {hdr}, DELIMITER {delim})"
                ).format(
                    tmp=tmp,
                    cols=col_idents,
                    hdr=psql.Literal("true" if csv_has_header else "false"),
                    delim=psql.Literal(csv_delimiter),
                )

                fh, should_close = _open_maybe_gz_read_text(src_path)
                try:
                    async with cur.copy(copy_sql) as cp:
                        async for chunk in _aiter_text_chunks(fh):
                            await cp.write(chunk)
                finally:
                    if should_close:
                        try:
                            fh.close()
                        except Exception:
                            pass

                # Upsert from staging into target
                insert_sql = (
                    psql.SQL("INSERT INTO {t} ({cols}) ")
                    + psql.SQL("SELECT {cols} FROM {tmp} ")
                    + psql.SQL("ON CONFLICT ({conflict}) DO UPDATE SET {set_list}")
                ).format(
                    t=psql.Identifier(target),
                    cols=col_idents,
                    tmp=tmp,
                    conflict=conflict_idents,
                    set_list=set_list,
                )

                await cur.execute(insert_sql)
                affected = cur.rowcount or 0

                # Keep planner fresh after heavy loads
                await cur.execute(psql.SQL("ANALYZE {t}").format(t=psql.Identifier(target)))

        return affected



"""
High-level write coordinator with metrics, health checks, and circuit breaker.
"""

from __future__ import annotations

import asyncio
import time
from dataclasses import dataclass
from typing import Generic, TypeVar, Sequence, Optional, Callable, Awaitable

from loguru import logger

from .policy import RetryPolicy, CircuitBreaker
from .queue import BoundedQueue, OverflowStrategy
from .types import Sink, BackpressureCallback
from .worker import SinkWorker
from .metrics import (
    COORD_ITEMS_SUBMITTED,
    COORD_ITEMS_DROPPED,
    COORD_QUEUE_DEPTH,
    COORD_WORKERS_ALIVE,
    CB_STATE,
)

T = TypeVar("T")


@dataclass
class CoordinatorHealth:
    """Health status of the coordinator."""

    workers_alive: int
    queue_size: int
    capacity: int
    circuit_state: str


class WriteCoordinator(Generic[T]):
    """High-level coordinator for producing → queue → workers → sink."""

    def __init__(
        self,
        *,
        sink: Sink[T],
        capacity: int = 10_000,
        workers: int = 4,
        batch_size: int = 500,
        flush_interval: float = 0.25,
        high_watermark: Optional[int] = None,
        low_watermark: Optional[int] = None,
        overflow_strategy: OverflowStrategy = "block",
        on_backpressure_high: Optional[BackpressureCallback] = None,
        on_backpressure_low: Optional[BackpressureCallback] = None,
        drop_callback: Optional[Callable[[T], Awaitable[None]]] = None,
        retry_policy: Optional[RetryPolicy] = None,
        circuit_breaker: Optional[CircuitBreaker] = None,
        coord_id: str = "default",
        metrics_poll_sec: float = 0.25,
    ):
        if workers <= 0:
            raise ValueError("workers must be > 0")
        if capacity <= 0:
            raise ValueError("capacity must be > 0")

        self._sink = sink
        self._coord_id = coord_id
        self._metrics_poll_sec = metrics_poll_sec
        self._cb = circuit_breaker or CircuitBreaker()

        # Wrap user-provided drop callback to also count metric
        async def _drop_with_metric(item: T) -> None:
            COORD_ITEMS_DROPPED.labels(self._coord_id, "overflow").inc()
            if drop_callback:
                await drop_callback(item)

        self._q = BoundedQueue[T](
            capacity=capacity,
            high_watermark=high_watermark,
            low_watermark=low_watermark,
            coord_id=coord_id,
            overflow_strategy=overflow_strategy,
            on_high=on_backpressure_high,
            on_low=on_backpressure_low,
            drop_callback=(_drop_with_metric if overflow_strategy == "drop_oldest" else None),
        )

        self._workers = [
            SinkWorker[T](
                worker_id=i,
                queue=self._q,
                sink=self._sink,
                batch_size=batch_size,
                flush_interval=flush_interval,
                retry_policy=retry_policy or RetryPolicy(),
                circuit_breaker=self._cb,
                coord_id=self._coord_id,
            )
            for i in range(workers)
        ]

        self._started = False
        self._metrics_task: Optional[asyncio.Task] = None

    async def __aenter__(self) -> "WriteCoordinator[T]":
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        await self.stop(drain=True)

    async def start(self) -> None:
        if self._started:
            return
        for w in self._workers:
            w.start()
        self._started = True
        self._metrics_task = asyncio.create_task(
            self._metrics_loop(), name=f"coord-metrics-{self._coord_id}"
        )
        logger.info(
            f"[coordinator {self._coord_id}] started "
            f"(cap={self._q.capacity}, workers={len(self._workers)})"
        )

    async def stop(self, drain: bool = True, timeout: float = 10.0) -> None:
        if not self._started:
            return

        if drain:
            start = time.perf_counter()
            while self._q.size > 0 and (time.perf_counter() - start) < timeout:
                await asyncio.sleep(0.05)
            if self._q.size > 0:
                logger.warning("[coordinator] drain timeout reached, forcing shutdown")

        # stop metrics first
        if self._metrics_task:
            self._metrics_task.cancel()
            try:
                await self._metrics_task
            except asyncio.CancelledError:
                pass

        await asyncio.gather(*(w.stop() for w in self._workers), return_exceptions=True)
        self._started = False
        logger.info(f"[coordinator {self._coord_id}] stopped")

    async def submit(self, item: T) -> None:
        """Submit a single item (observes overflow strategy)."""
        COORD_ITEMS_SUBMITTED.labels(self._coord_id).inc()
        await self._q.put(item)

    async def submit_many(self, items: Sequence[T]) -> None:
        """Submit multiple items."""
        for it in items:
            await self.submit(it)

    def health(self) -> CoordinatorHealth:
        """Get current health status."""
        # workers alive
        alive = 0
        for w in self._workers:
            t = getattr(w, "_task", None)
            if t and not t.done():
                alive += 1
        state = self._cb.state
        return CoordinatorHealth(
            workers_alive=alive,
            queue_size=self._q.size,
            capacity=self._q.capacity,
            circuit_state=state,
        )

    async def _metrics_loop(self) -> None:
        """Background task to update Prometheus gauges."""
        try:
            while True:
                COORD_QUEUE_DEPTH.labels(self._coord_id).set(self._q.size)
                # map CB state
                state_map = {"closed": 0, "open": 1, "half_open": 2}
                CB_STATE.labels(self._coord_id).set(state_map.get(self._cb.state, -1))
                COORD_WORKERS_ALIVE.labels(self._coord_id).set(
                    sum(
                        1
                        for w in self._workers
                        if getattr(w, "_task", None) and not w._task.done()  # noqa: SLF001
                    )
                )
                await asyncio.sleep(self._metrics_poll_sec)
        except asyncio.CancelledError:
            return



from .client import MDS, MDSConfig
from .aclient import AMDS, AMDSConfig  # includes copy_out_ndjson_async, copy_restore_csv_async
from .sql import TABLE_PRESETS, build_ndjson_select
from .models import Bar, Fundamentals, News, OptionSnap, LatestPrice
from .batch import BatchProcessor, AsyncBatchProcessor, BatchConfig

__all__ = [
    "MDS",
    "MDSConfig",
    "AMDS",
    "AMDSConfig",
    "TABLE_PRESETS",
    "build_ndjson_select",
    "Bar",
    "Fundamentals",
    "News",
    "OptionSnap",
    "LatestPrice",
    "BatchProcessor",
    "AsyncBatchProcessor",
    "BatchConfig",
]



from __future__ import annotations

import asyncio
import json
import os
import sys
import tempfile
from datetime import datetime
from pathlib import Path
from typing import Optional

import typer

from . import MDS, AMDS
from .batch import BatchProcessor, AsyncBatchProcessor, BatchConfig
from .models import Bar, Fundamentals, News, OptionSnap
from .utils import iter_ndjson, coerce_model
from .sql import TABLE_PRESETS
from .runtime import boot_event_loop, shutdown_with_timeout
from .health import (
    check_health,
    check_health_with_retry,
    get_prometheus_metrics,
    get_metrics_summary,
)

app = typer.Typer(help="mds_client operational CLI")

# ---------------------------
# Common options
# ---------------------------


def dsn_opt() -> str:
    return typer.Option(..., "--dsn", envvar="MDS_DSN", help="PostgreSQL DSN")


def tenant_opt() -> str:
    return typer.Option(
        ...,
        "--tenant-id",
        envvar="MDS_TENANT_ID",
        help="Tenant UUID for RLS (tenants.id, not tenants.tenant_id)",
    )


def vendor_opt() -> Optional[str]:
    return typer.Option(None, "--vendor", help="Data vendor (e.g. ibkr, reuters)")


def max_rows_opt(default=1000) -> int:
    return typer.Option(default, "--max-rows", help="Flush when pending rows reach this size")


def max_ms_opt(default=5000) -> int:
    return typer.Option(default, "--max-ms", help="Flush when this many ms elapse since last flush")


def max_bytes_opt(default=1_048_576) -> int:
    return typer.Option(default, "--max-bytes", help="Flush when pending bytes reach this size")


# ---------------------------
# Health / Schema / Reads
# ---------------------------


@app.command("ping")
def ping(dsn: str = dsn_opt(), tenant_id: str = tenant_opt()):
    mds = MDS({"dsn": dsn, "tenant_id": tenant_id})
    ok = mds.health()
    typer.echo(json.dumps({"ok": ok}, indent=2))


@app.command("schema-version")
def schema_version(dsn: str = dsn_opt(), tenant_id: str = tenant_opt()):
    mds = MDS({"dsn": dsn, "tenant_id": tenant_id})
    ver = mds.schema_version()
    typer.echo(json.dumps({"schema_version": ver}, indent=2))


@app.command("health")
def health_check(
    dsn: str = dsn_opt(),
    tenant_id: str = tenant_opt(),
    retry: bool = typer.Option(False, "--retry", help="Use retry logic for health check"),
    format: str = typer.Option("json", "--format", help="Output format: json, prometheus"),
):
    """Comprehensive database health check with metrics."""

    async def _health_check():
        amds = AMDS({"dsn": dsn, "tenant_id": tenant_id, "pool_max": 5})
        try:
            await amds.aopen()

            if retry:
                result = await check_health_with_retry(amds)
            else:
                result = await check_health(amds)

            if format == "prometheus":
                metrics = get_prometheus_metrics()
                if metrics:
                    typer.echo(metrics)
                else:
                    typer.echo(
                        "# Prometheus metrics not available (prometheus_client not installed)"
                    )
            else:
                typer.echo(json.dumps(result, indent=2, default=str))

        finally:
            await amds.aclose()

    asyncio.run(_health_check())


@app.command("metrics")
def metrics(
    format: str = typer.Option("json", "--format", help="Output format: json, prometheus"),
):
    """Get current metrics summary."""
    if format == "prometheus":
        metrics = get_prometheus_metrics()
        if metrics:
            typer.echo(metrics)
        else:
            typer.echo("# Prometheus metrics not available (prometheus_client not installed)")
    else:
        summary = get_metrics_summary()
        typer.echo(json.dumps(summary, indent=2, default=str))


@app.command("latest-prices")
def latest_prices(
    symbols: str = typer.Argument(..., help="Comma-separated symbols"),
    vendor: str = typer.Option(..., "--vendor", help="Data vendor"),
    dsn: str = dsn_opt(),
    tenant_id: str = tenant_opt(),
):
    mds = MDS({"dsn": dsn, "tenant_id": tenant_id})
    syms = [s.strip().upper() for s in symbols.split(",") if s.strip()]
    rows = mds.latest_prices(syms, vendor=vendor)
    for r in rows:
        typer.echo(json.dumps(r, default=str))


# ---------------------------
# Write commands (sync)
# ---------------------------


@app.command("write-bar")
def write_bar(
    symbol: str = typer.Option(...),
    timeframe: str = typer.Option(...),
    ts: datetime = typer.Option(..., formats=["%Y-%m-%dT%H:%M:%S", "%Y-%m-%dT%H:%M:%S%z"]),
    close_price: Optional[float] = typer.Option(None),
    open_price: Optional[float] = typer.Option(None),
    high_price: Optional[float] = typer.Option(None),
    low_price: Optional[float] = typer.Option(None),
    volume: Optional[int] = typer.Option(None),
    vendor: str = typer.Option(..., "--vendor"),
    dsn: str = dsn_opt(),
    tenant_id: str = tenant_opt(),
):
    mds = MDS({"dsn": dsn, "tenant_id": tenant_id})
    bar = Bar(
        tenant_id=tenant_id,
        vendor=vendor,
        symbol=symbol,
        timeframe=timeframe,
        ts=ts,
        open_price=open_price,
        high_price=high_price,
        low_price=low_price,
        close_price=close_price,
        volume=volume,
    )
    mds.upsert_bars([bar])
    typer.echo("ok")


@app.command("write-fundamental")
def write_fundamental(
    symbol: str = typer.Option(...),
    asof: datetime = typer.Option(
        ..., formats=["%Y-%m-%d", "%Y-%m-%dT%H:%M:%S", "%Y-%m-%dT%H:%M:%S%z"]
    ),
    total_assets: Optional[float] = typer.Option(None),
    total_liabilities: Optional[float] = typer.Option(None),
    net_income: Optional[float] = typer.Option(None),
    eps: Optional[float] = typer.Option(None),
    vendor: str = typer.Option(..., "--vendor"),
    dsn: str = dsn_opt(),
    tenant_id: str = tenant_opt(),
):
    mds = MDS({"dsn": dsn, "tenant_id": tenant_id})
    row = Fundamentals(
        tenant_id=tenant_id,
        vendor=vendor,
        symbol=symbol,
        asof=asof,
        total_assets=total_assets,
        total_liabilities=total_liabilities,
        net_income=net_income,
        eps=eps,
    )
    mds.upsert_fundamentals([row])
    typer.echo("ok")


@app.command("write-news")
def write_news(
    title: str = typer.Option(...),
    published_at: datetime = typer.Option(
        ..., formats=["%Y-%m-%dT%H:%M:%S", "%Y-%m-%dT%H:%M:%S%z"]
    ),
    symbol: Optional[str] = typer.Option(None),
    url: Optional[str] = typer.Option(None),
    sentiment_score: Optional[float] = typer.Option(None),
    vendor: str = typer.Option(..., "--vendor"),
    dsn: str = dsn_opt(),
    tenant_id: str = tenant_opt(),
):
    mds = MDS({"dsn": dsn, "tenant_id": tenant_id})
    row = News(
        tenant_id=tenant_id,
        vendor=vendor,
        title=title,
        published_at=published_at,
        symbol=symbol,
        url=url,
        sentiment_score=sentiment_score,
    )
    mds.upsert_news([row])
    typer.echo("ok")


@app.command("write-option")
def write_option(
    symbol: str = typer.Option(...),
    expiry: str = typer.Option(..., help="YYYY-MM-DD"),
    option_type: str = typer.Option(..., help="'C' or 'P'"),
    strike: float = typer.Option(...),
    ts: datetime = typer.Option(..., formats=["%Y-%m-%dT%H:%M:%S", "%Y-%m-%dT%H:%M:%S%z"]),
    iv: Optional[float] = typer.Option(None),
    delta: Optional[float] = typer.Option(None),
    gamma: Optional[float] = typer.Option(None),
    oi: Optional[int] = typer.Option(None),
    volume: Optional[int] = typer.Option(None),
    spot: Optional[float] = typer.Option(None),
    vendor: str = typer.Option(..., "--vendor"),
    dsn: str = dsn_opt(),
    tenant_id: str = tenant_opt(),
):
    from datetime import date

    y, m, d = [int(x) for x in expiry.split("-")]
    mds = MDS({"dsn": dsn, "tenant_id": tenant_id})
    row = OptionSnap(
        tenant_id=tenant_id,
        vendor=vendor,
        symbol=symbol,
        expiry=date(y, m, d),
        option_type=option_type,
        strike=strike,
        ts=ts,
        iv=iv,
        delta=delta,
        gamma=gamma,
        oi=oi,
        volume=volume,
        spot=spot,
    )
    mds.upsert_options([row])
    typer.echo("ok")


# ---------------------------
# NDJSON ingest
# ---------------------------


@app.command("ingest-ndjson")
def ingest_ndjson(
    kind: str = typer.Argument(..., help="bars|fundamentals|news|options"),
    path: str = typer.Argument(..., help="File path or '-' for stdin (.gz ok)"),
    dsn: str = dsn_opt(),
    tenant_id: str = tenant_opt(),
    max_rows: int = max_rows_opt(1000),
    max_ms: int = max_ms_opt(5000),
    max_bytes: int = max_bytes_opt(1_048_576),
):
    kind_l = kind.lower()
    if kind_l not in ("bars", "fundamentals", "news", "options"):
        raise typer.BadParameter("kind must be one of: bars, fundamentals, news, options")

    cfg = BatchConfig(max_rows=max_rows, max_ms=max_ms, max_bytes=max_bytes)
    mds = MDS({"dsn": dsn, "tenant_id": tenant_id})
    bp = BatchProcessor(mds, cfg)

    add_fn = {
        "bars": bp.add_bar,
        "fundamentals": bp.add_fundamental,
        "news": bp.add_news,
        "options": bp.add_option,
    }[kind_l]

    n = 0
    for obj in iter_ndjson(path):
        row = coerce_model(kind_l, obj)
        add_fn(row)
        n += 1

    counts = bp.flush()
    typer.echo(json.dumps({"ingested": n, "flushed": counts}, default=str, indent=2))


@app.command("ingest-ndjson-async")
def ingest_ndjson_async(
    kind: str = typer.Argument(..., help="bars|fundamentals|news|options"),
    path: str = typer.Argument(...),
    dsn: str = dsn_opt(),
    tenant_id: str = tenant_opt(),
    max_rows: int = max_rows_opt(1000),
    max_ms: int = max_ms_opt(5000),
    max_bytes: int = max_bytes_opt(1_048_576),
):
    asyncio.run(_ingest_ndjson_async(kind, path, dsn, tenant_id, max_rows, max_ms, max_bytes))


async def _ingest_ndjson_async(
    kind: str, path: str, dsn: str, tenant_id: str, max_rows: int, max_ms: int, max_bytes: int
):
    kind_l = kind.lower()
    if kind_l not in ("bars", "fundamentals", "news", "options"):
        raise typer.BadParameter("kind must be one of: bars, fundamentals, news, options")

    cfg = BatchConfig(max_rows=max_rows, max_ms=max_ms, max_bytes=max_bytes)
    amds = AMDS({"dsn": dsn, "tenant_id": tenant_id, "pool_max": 10})

    try:
        async with AsyncBatchProcessor(amds, cfg) as bp:
            add_fn = {
                "bars": bp.add_bar,
                "fundamentals": bp.add_fundamental,
                "news": bp.add_news,
                "options": bp.add_option,
            }[kind_l]
            n = 0
            for obj in iter_ndjson(path):
                await add_fn(coerce_model(kind_l, obj))
                n += 1
        # Auto-flush on exit
        typer.echo(json.dumps({"ingested": n, "flushed": "auto"}, default=str, indent=2))
    finally:
        # Ensure proper cleanup
        await shutdown_with_timeout(amds.pool)


# ---------------------------
# Jobs outbox (simple helper)
# ---------------------------


@app.command("enqueue-job")
def enqueue_job(
    idempotency_key: str = typer.Option(..., "--idempotency-key"),
    job_type: str = typer.Option(..., "--job-type"),
    payload: str = typer.Option(..., "--payload", help="JSON string"),
    priority: str = typer.Option("medium", "--priority"),
    dsn: str = dsn_opt(),
    tenant_id: str = tenant_opt(),
):
    """Minimal helper that inserts into jobs_outbox with conflict-free idempotency."""
    mds = MDS({"dsn": dsn, "tenant_id": tenant_id})
    mds.enqueue_job(
        idempotency_key=idempotency_key,
        job_type=job_type,
        payload=json.loads(payload),
        priority=priority,
    )
    typer.echo("ok")


# ---------------------------
# Backup/Export/Import commands
# ---------------------------


@app.command("dump")
def dump(
    table: str = typer.Argument(..., help="bars|fundamentals|news|options_snap"),
    out_path: str = typer.Argument(..., help="Output file (.csv or .csv.gz) or '-' for stdout"),
    dsn: str = dsn_opt(),
    tenant_id: str = tenant_opt(),
    vendor: str = typer.Option(None, "--vendor", help="Filter by vendor"),
    symbol: str = typer.Option(None, "--symbol", help="Filter by symbol"),
    timeframe: str = typer.Option(None, "--timeframe", help="Filter by timeframe"),
    start: str = typer.Option(None, "--start", help="Start time (ISO format)"),
    end: str = typer.Option(None, "--end", help="End time (ISO format)"),
):
    """Export table data to CSV with optional filters."""
    if table not in TABLE_PRESETS:
        raise typer.BadParameter(f"table must be one of: {', '.join(TABLE_PRESETS.keys())}")

    mds = MDS({"dsn": dsn, "tenant_id": tenant_id})
    sel = mds.build_ndjson_select(
        table, vendor=vendor, symbol=symbol, timeframe=timeframe, start=start, end=end
    )
    nbytes = mds.copy_out_csv(select_sql=sel, out_path=out_path)
    typer.echo(f"wrote {nbytes} bytes")


@app.command("restore")
def restore(
    table: str = typer.Argument(..., help="bars|fundamentals|news|options_snap"),
    src_path: str = typer.Argument(..., help="Input file (.csv or .csv.gz)"),
    dsn: str = dsn_opt(),
    tenant_id: str = tenant_opt(),
):
    """Import table data from CSV with upsert semantics."""
    if table not in TABLE_PRESETS:
        raise typer.BadParameter(f"table must be one of: {', '.join(TABLE_PRESETS.keys())}")

    mds = MDS({"dsn": dsn, "tenant_id": tenant_id})
    p = TABLE_PRESETS[table]
    n = mds.copy_restore_csv(
        target=table,
        cols=p["cols"],
        conflict_cols=p["conflict"],
        update_cols=p["update"],
        src_path=src_path,
    )
    typer.echo(f"upserted {n} rows")


# --- async restore CSV -------------------------------------------------------


async def _restore_csv_async_impl(
    table: str,
    src_path: Path,
    dsn: str,
    tenant_id: Optional[str],
    app_name: str,
    pool_max: int,
    delimiter: str,
    header: bool,
    null: str,
    temp_table_name: Optional[str],
) -> int:
    if table not in TABLE_PRESETS:
        typer.secho(
            f"Unknown table '{table}'. Valid: {', '.join(TABLE_PRESETS)}",
            fg=typer.colors.RED,
            err=True,
        )
        raise typer.Exit(2)

    amds = AMDS(
        {
            "dsn": dsn,
            "tenant_id": tenant_id,
            "app_name": app_name,
            "pool_max": pool_max,
        }
    )
    preset = TABLE_PRESETS[table]
    try:
        n = await amds.copy_restore_csv_async(
            target=table,
            cols=preset.cols,
            conflict_cols=preset.conflict,
            update_cols=preset.update,
            src_path=str(src_path),
            csv_delimiter=delimiter,
            csv_has_header=header,
        )
        return n
    finally:
        await amds.aclose()


@app.command("restore-async")
def restore_async(
    table: str = typer.Argument(..., help="Target table: bars|fundamentals|news|options_snap"),
    src_path: Path = typer.Argument(
        ..., exists=True, readable=True, help="CSV or CSV.GZ to restore"
    ),
    dsn: str = typer.Option(None, "--dsn", envvar="MDS_DSN", help="PostgreSQL DSN"),
    tenant_id: Optional[str] = typer.Option(
        None, "--tenant-id", envvar="MDS_TENANT_ID", help="Tenant UUID for RLS"
    ),
    app_name: str = typer.Option("mds_client_async", "--app-name", help="application_name for pg"),
    pool_max: int = typer.Option(10, "--pool-max", min=1, help="Async pool max connections"),
    delimiter: str = typer.Option(",", "--delimiter", help="CSV delimiter"),
    header: bool = typer.Option(True, "--header/--no-header", help="CSV has header row"),
    null: str = typer.Option("\\N", "--null", help="NULL representation"),
    temp_table_name: Optional[str] = typer.Option(
        None, "--temp-table", help="Override temp staging table name"
    ),
):
    """
    Async CSV restore with idempotent upserts (COPY to staging → INSERT ... ON CONFLICT).
    """
    if not dsn:
        typer.secho("Missing DSN. Provide --dsn or set MDS_DSN.", fg=typer.colors.RED, err=True)
        raise typer.Exit(2)

    n = asyncio.run(
        _restore_csv_async_impl(
            table=table,
            src_path=src_path,
            dsn=dsn,
            tenant_id=tenant_id,
            app_name=app_name,
            pool_max=pool_max,
            delimiter=delimiter,
            header=header,
            null=null,
            temp_table_name=temp_table_name,
        )
    )
    typer.secho(f"✅ restore-async complete: {n} rows upserted into {table}", fg=typer.colors.GREEN)


# Optional alias with a more explicit name
@app.command("restore-csv-async")
def restore_csv_async_alias(
    table: str = typer.Argument(...),
    src_path: Path = typer.Argument(...),
    dsn: str = typer.Option(None, "--dsn", envvar="MDS_DSN"),
    tenant_id: Optional[str] = typer.Option(None, "--tenant-id", envvar="MDS_TENANT_ID"),
    app_name: str = typer.Option("mds_client_async", "--app-name"),
    pool_max: int = typer.Option(10, "--pool-max", min=1),
    delimiter: str = typer.Option(",", "--delimiter"),
    header: bool = typer.Option(True, "--header/--no-header"),
    null: str = typer.Option("\\N", "--null"),
    temp_table_name: Optional[str] = typer.Option(None, "--temp-table"),
):
    """Alias of restore-async."""
    return restore_async(
        table=table,
        src_path=src_path,
        dsn=dsn,
        tenant_id=tenant_id,
        app_name=app_name,
        pool_max=pool_max,
        delimiter=delimiter,
        header=header,
        null=null,
        temp_table_name=temp_table_name,
    )


# --- async restore CSV from STDIN -------------------------------------------


@app.command("restore-async-stdin")
def restore_async_stdin(
    table: str = typer.Argument(..., help="Target table: bars|fundamentals|news|options_snap"),
    dsn: str = typer.Option(None, "--dsn", envvar="MDS_DSN", help="PostgreSQL DSN"),
    tenant_id: Optional[str] = typer.Option(
        None, "--tenant-id", envvar="MDS_TENANT_ID", help="Tenant UUID for RLS"
    ),
    app_name: str = typer.Option("mds_client_async", "--app-name", help="application_name for pg"),
    pool_max: int = typer.Option(10, "--pool-max", min=1, help="Async pool max connections"),
    delimiter: str = typer.Option(",", "--delimiter", help="CSV delimiter"),
    header: bool = typer.Option(True, "--header/--no-header", help="CSV has header row"),
    null: str = typer.Option("\\N", "--null", help="NULL representation"),
    temp_table_name: Optional[str] = typer.Option(
        None, "--temp-table", help="Override temp staging table name"
    ),
):
    """
    Async CSV restore from STDIN with idempotent upserts (COPY to staging → INSERT ... ON CONFLICT).
    Example:
      zcat bars.csv.gz | mds restore-async-stdin bars --dsn ... --tenant-id ...
    """
    if not dsn:
        typer.secho("Missing DSN. Provide --dsn or set MDS_DSN.", fg=typer.colors.RED, err=True)
        raise typer.Exit(2)

    # stream stdin to a temp file on disk (works cross-platform)
    tmp_path: Optional[Path] = None
    try:
        with tempfile.NamedTemporaryFile(mode="wb", suffix=".csv", delete=False) as tmp:
            tmp_path = Path(tmp.name)
            buf = sys.stdin.buffer
            chunk = buf.read(1 << 20)  # 1 MiB chunks
            while chunk:
                tmp.write(chunk)
                chunk = buf.read(1 << 20)

        # reuse the existing async CSV restore impl (treat temp file as plain CSV)
        n = asyncio.run(
            _restore_csv_async_impl(
                table=table,
                src_path=tmp_path,
                dsn=dsn,
                tenant_id=tenant_id,
                app_name=app_name,
                pool_max=pool_max,
                delimiter=delimiter,
                header=header,
                null=null,
                temp_table_name=temp_table_name,
            )
        )
        typer.secho(
            f"✅ restore-async (stdin) complete: {n} rows upserted into {table}",
            fg=typer.colors.GREEN,
        )
    finally:
        if tmp_path:
            try:
                os.remove(tmp_path)
            except OSError:
                pass


# --- NDJSON export commands (dump-ndjson*) -----------------------------------

# ---------- small helpers ----------


def _env(k: str, default: Optional[str] = None) -> Optional[str]:
    v = os.getenv(k)
    return v if v not in (None, "") else default


def _cfg(dsn: Optional[str], tenant_id: Optional[str], *, async_mode: bool = False):
    dsn = dsn or _env("MDS_DSN")
    tenant_id = tenant_id or _env("MDS_TENANT_ID")
    if not dsn:
        raise typer.BadParameter("Missing DSN. Provide --dsn or set MDS_DSN.")
    if not tenant_id:
        raise typer.BadParameter("Missing tenant id. Provide --tenant-id or set MDS_TENANT_ID.")
    base = {
        "dsn": dsn,
        "tenant_id": tenant_id,
        "app_name": "mds_client_async" if async_mode else "mds_client",
    }
    return base


def _ensure_parent(path: Path) -> None:
    if path != Path("-"):
        path.parent.mkdir(parents=True, exist_ok=True)


def _slug(v: Optional[str], fallback: str) -> str:
    if not v:
        return fallback
    return str(v).replace(":", "").replace("/", "_").replace("\\", "_").replace(" ", "")


def _format_template(
    template: str,
    *,
    table: str,
    vendor: Optional[str],
    symbol: Optional[str],
    timeframe: Optional[str],
    start: Optional[str],
    end: Optional[str],
) -> str:
    return template.format(
        table=table,
        vendor=_slug(vendor, "ALL"),
        symbol=_slug(symbol, "ALL"),
        timeframe=_slug(timeframe, "ALL"),
        start=_slug(start, "MIN"),
        end=_slug(end, "MAX"),
    )


# ---------- single-table sync ----------


@app.command("dump-ndjson")
def dump_ndjson(
    table: str = typer.Argument(..., help=f"Table name ({', '.join(TABLE_PRESETS.keys())})"),
    out_path: Path = typer.Argument(..., help="Output file path (.ndjson or .ndjson.gz)"),
    dsn: Optional[str] = typer.Option(None, help="PostgreSQL DSN (or set MDS_DSN)"),
    tenant_id: Optional[str] = typer.Option(
        None, "--tenant-id", help="Tenant UUID (or set MDS_TENANT_ID)"
    ),
    vendor: Optional[str] = typer.Option(None, help="Filter: vendor"),
    symbol: Optional[str] = typer.Option(None, help="Filter: symbol"),
    timeframe: Optional[str] = typer.Option(None, help="Filter: timeframe"),
    start: Optional[str] = typer.Option(None, help="Filter: start timestamp (ISO-8601)"),
    end: Optional[str] = typer.Option(None, help="Filter: end timestamp (ISO-8601)"),
):
    """
    Export a table to NDJSON using COPY (round-trips with ingest-ndjson).
    """
    if table not in TABLE_PRESETS:
        raise typer.BadParameter(
            f"Unknown table '{table}'. Valid: {', '.join(TABLE_PRESETS.keys())}"
        )

    cfg = _cfg(dsn, tenant_id, async_mode=False)
    mds = MDS(cfg)
    try:
        sel = mds.build_ndjson_select(
            table,
            vendor=vendor,
            symbol=symbol,
            timeframe=timeframe,
            start=start,
            end=end,
        )
        _ensure_parent(out_path)
        nbytes = mds.copy_out_ndjson(select_sql=sel, out_path=str(out_path))
        typer.echo(f"Wrote {nbytes} bytes → {out_path}")
    finally:
        mds.close()


# ---------- single-table async ----------


@app.command("dump-ndjson-async")
def dump_ndjson_async(
    table: str = typer.Argument(..., help=f"Table name ({', '.join(TABLE_PRESETS.keys())})"),
    out_path: Path = typer.Argument(..., help="Output file path (.ndjson or .ndjson.gz)"),
    dsn: Optional[str] = typer.Option(None, help="PostgreSQL DSN (or set MDS_DSN)"),
    tenant_id: Optional[str] = typer.Option(
        None, "--tenant-id", help="Tenant UUID (or set MDS_TENANT_ID)"
    ),
    vendor: Optional[str] = typer.Option(None, help="Filter: vendor"),
    symbol: Optional[str] = typer.Option(None, help="Filter: symbol"),
    timeframe: Optional[str] = typer.Option(None, help="Filter: timeframe"),
    start: Optional[str] = typer.Option(None, help="Filter: start timestamp (ISO-8601)"),
    end: Optional[str] = typer.Option(None, help="Filter: end timestamp (ISO-8601)"),
):
    """
    Async export of a table to NDJSON using COPY.
    """
    if table not in TABLE_PRESETS:
        raise typer.BadParameter(
            f"Unknown table '{table}'. Valid: {', '.join(TABLE_PRESETS.keys())}"
        )

    async def _run():
        cfg = _cfg(dsn, tenant_id, async_mode=True)
        amds = AMDS(cfg)
        try:
            from .sql import build_ndjson_select as _build

            preset = TABLE_PRESETS[table]
            sel = _build(
                table,
                preset["cols"],
                vendor=vendor,
                symbol=(symbol.upper() if symbol else None),
                timeframe=timeframe,
                start=start,
                end=end,
            )
            _ensure_parent(out_path)
            nbytes = await amds.copy_out_ndjson_async(select_sql=sel, out_path=str(out_path))
            typer.echo(f"Wrote {nbytes} bytes → {out_path}")
        finally:
            await amds.aclose()

    asyncio.run(_run())


# ---------- multi-table sync ----------


@app.command("dump-ndjson-all")
def dump_ndjson_all(
    template: str = typer.Argument(
        "{table}-{vendor}-{symbol}-{start}-{end}.ndjson.gz",
        help="Output name template with placeholders: {table},{vendor},{symbol},{timeframe},{start},{end}",
    ),
    dsn: Optional[str] = typer.Option(None, help="PostgreSQL DSN (or set MDS_DSN)"),
    tenant_id: Optional[str] = typer.Option(
        None, "--tenant-id", help="Tenant UUID (or set MDS_TENANT_ID)"
    ),
    vendor: Optional[str] = typer.Option(None, help="Filter: vendor"),
    symbol: Optional[str] = typer.Option(None, help="Filter: symbol"),
    timeframe: Optional[str] = typer.Option(None, help="Filter: timeframe"),
    start: Optional[str] = typer.Option(None, help="Filter: start timestamp (ISO-8601)"),
    end: Optional[str] = typer.Option(None, help="Filter: end timestamp (ISO-8601)"),
):
    """
    Export ALL known tables to NDJSON files using a naming template.
    """
    cfg = _cfg(dsn, tenant_id, async_mode=False)
    mds = MDS(cfg)
    try:
        for table in TABLE_PRESETS.keys():
            sel = mds.build_ndjson_select(
                table,
                vendor=vendor,
                symbol=symbol,
                timeframe=timeframe,
                start=start,
                end=end,
            )
            out_name = _format_template(
                template,
                table=table,
                vendor=vendor,
                symbol=symbol,
                timeframe=timeframe,
                start=start,
                end=end,
            )
            out_path = Path(out_name)
            _ensure_parent(out_path)
            nbytes = mds.copy_out_ndjson(select_sql=sel, out_path=str(out_path))
            typer.echo(f"[{table}] wrote {nbytes} bytes → {out_path}")
    finally:
        mds.close()


# ---------- multi-table async ----------


@app.command("dump-ndjson-async-all")
def dump_ndjson_async_all(
    template: str = typer.Argument(
        "{table}-{vendor}-{symbol}-{start}-{end}.ndjson.gz",
        help="Output name template with placeholders: {table},{vendor},{symbol},{timeframe},{start},{end}",
    ),
    dsn: Optional[str] = typer.Option(None, help="PostgreSQL DSN (or set MDS_DSN)"),
    tenant_id: Optional[str] = typer.Option(
        None, "--tenant-id", help="Tenant UUID (or set MDS_TENANT_ID)"
    ),
    vendor: Optional[str] = typer.Option(None, help="Filter: vendor"),
    symbol: Optional[str] = typer.Option(None, help="Filter: symbol"),
    timeframe: Optional[str] = typer.Option(None, help="Filter: timeframe"),
    start: Optional[str] = typer.Option(None, help="Filter: start timestamp (ISO-8601)"),
    end: Optional[str] = typer.Option(None, help="Filter: end timestamp (ISO-8601)"),
):
    """
    Async export of ALL tables to NDJSON files using a naming template.
    """

    async def _run():
        cfg = _cfg(dsn, tenant_id, async_mode=True)
        amds = AMDS(cfg)
        try:
            for table in TABLE_PRESETS.keys():
                # Use the shared builder from sql.py directly to avoid any attribute coupling
                from .sql import build_ndjson_select as _build

                preset = TABLE_PRESETS[table]
                sel = _build(
                    table,
                    preset["cols"],
                    vendor=vendor,
                    symbol=(symbol.upper() if symbol else None),
                    timeframe=timeframe,
                    start=start,
                    end=end,
                )
                out_name = _format_template(
                    template,
                    table=table,
                    vendor=vendor,
                    symbol=symbol,
                    timeframe=timeframe,
                    start=start,
                    end=end,
                )
                out_path = Path(out_name)
                _ensure_parent(out_path)
                nbytes = await amds.copy_out_ndjson_async(select_sql=sel, out_path=str(out_path))
                typer.echo(f"[{table}] wrote {nbytes} bytes → {out_path}")
        finally:
            await amds.aclose()

    asyncio.run(_run())


# Configure event loop on module import
boot_event_loop()

# If this module is run directly:
if __name__ == "__main__":  # pragma: no cover
    app()
